/*
 * This is free and unencumbered software released into the public domain.
 *
 * Anyone is free to copy, modify, publish, use, compile, sell, or
 * distribute this software, either in source code form or as a compiled
 * binary, for any purpose, commercial or non-commercial, and by any
 * means.
 *
 * In jurisdictions that recognize copyright laws, the author or authors
 * of this software dedicate any and all copyright interest in the
 * software to the public domain. We make this dedication for the benefit
 * of the public at large and to the detriment of our heirs and
 * successors. We intend this dedication to be an overt act of
 * relinquishment in perpetuity of all present and future rights to this
 * software under copyright law.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * For more information, please refer to <http://unlicense.org/>
 */

package org.jenkinsci.plugins.graniteclient;

import java.util.Collections;
import java.util.List;

import com.cloudbees.jenkins.plugins.sshcredentials.SSHUserPrivateKey;
import com.cloudbees.plugins.credentials.Credentials;
import com.cloudbees.plugins.credentials.CredentialsNameProvider;
import com.cloudbees.plugins.credentials.CredentialsProvider;
import com.cloudbees.plugins.credentials.CredentialsScope;
import com.cloudbees.plugins.credentials.common.AbstractIdCredentialsListBoxModel;
import com.cloudbees.plugins.credentials.common.IdCredentials;
import com.cloudbees.plugins.credentials.common.StandardUsernamePasswordCredentials;
import com.cloudbees.plugins.credentials.domains.DomainRequirement;
import com.cloudbees.plugins.credentials.domains.URIRequirementBuilder;
import edu.umd.cs.findbugs.annotations.NonNull;
import hudson.model.Item;
import hudson.model.Queue;
import hudson.model.queue.Tasks;
import hudson.security.ACL;
import hudson.security.AccessControlled;
import jenkins.model.Jenkins;

/**
 *
 */
public class GraniteCredentialsListBoxModel extends AbstractIdCredentialsListBoxModel<GraniteCredentialsListBoxModel, IdCredentials> {

    private static final long serialVersionUID = 6621529150670191091L;

    @NonNull
    @Override
    protected String describe(@NonNull IdCredentials idCredentials) {
        if (idCredentials instanceof GraniteNamedIdCredentials) {
            return ((GraniteNamedIdCredentials) idCredentials).getName();
        } else {
            return CredentialsNameProvider.name(idCredentials);
        }
    }

    public static AbstractIdCredentialsListBoxModel fillItems(final String currentValue, final Item context) {
        return fillItems(currentValue, context, Collections.<DomainRequirement>emptyList());
    }

    public static AbstractIdCredentialsListBoxModel fillItems(final String currentValue, final Item context, final String baseUrl) {
        if (baseUrl != null) {
            return fillItems(currentValue, context, URIRequirementBuilder.fromUri(baseUrl).build());
        } else {
            return fillItems(currentValue, context);
        }
    }

    private static AbstractIdCredentialsListBoxModel fillItems(final String currentValue, final Item context, final List<DomainRequirement> reqs) {
        AbstractIdCredentialsListBoxModel<GraniteCredentialsListBoxModel, IdCredentials> model =
                new GraniteCredentialsListBoxModel().withEmptySelection();

        // when a context is provided (by a job config) but lacks configure permissions
        // return the existing value
        if (context != null && !context.hasPermission(Item.CONFIGURE)) {
            Credentials _credentials = GraniteNamedIdCredentials.getCredentialsById(currentValue);
            return model.with(GraniteNamedIdCredentials.maybeWrap(_credentials));
        }

        // when a context is not provided (by global config) but lacks admin permissions
        // return the existing value
        if (context == null && !Jenkins.getActiveInstance().hasPermission(Jenkins.ADMINISTER)) {
            Credentials _credentials = GraniteNamedIdCredentials.getCredentialsById(currentValue);
            return model.with(GraniteNamedIdCredentials.maybeWrap(_credentials));
        }

        List<SSHUserPrivateKey> keys = CredentialsProvider.lookupCredentials(SSHUserPrivateKey.class,
                context,
                context instanceof Queue.Task ? Tasks.getAuthenticationOf((Queue.Task) context) : ACL.SYSTEM,
                reqs);

        if (!keys.isEmpty()) {
            for (SSHUserPrivateKey key : keys) {
                if (key.getScope() == CredentialsScope.GLOBAL) {
                    model = model.with(GraniteNamedIdCredentials.wrap(key));
                }
            }
        }

        List<StandardUsernamePasswordCredentials> basicAuthCredsList =
                CredentialsProvider.lookupCredentials(StandardUsernamePasswordCredentials.class,
                        context,
                        context instanceof Queue.Task ? Tasks.getAuthenticationOf((Queue.Task) context) : ACL.SYSTEM,
                        reqs);

        if (!basicAuthCredsList.isEmpty()) {
            for (StandardUsernamePasswordCredentials creds : basicAuthCredsList) {
                if (creds.getScope() == CredentialsScope.GLOBAL) {
                    model = model.with(GraniteNamedIdCredentials.wrap(creds));
                }
            }
        }

        return model;
    }
}
